/**
* @file Arabic Unicode to CAL Code conversion
* @version 1.0.4
* @author Greg Borota
* @copyright (c) 2017 Greg Borota.
* @license MIT
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*/

// https://peshitta.github.io
// https://sedra.bethmardutho.org/about/fonts
// http://cal1.cn.huc.edu/searching/fullbrowser.html

import { Mapper, Writing } from 'aramaic-mapper';
import { allConsonants, diacritics, isDotted, punctuation, vowels } from 'arabic-code-util';
import { consonants } from 'cal-code-util';

/** @module arabicCal */
/**
 * @private
 * Arabic source writing
 * @const
 * @type { Writing }
 */
var arabicWriting = new Writing(
  allConsonants,
  vowels,
  diacritics,
  punctuation
);

var thehTo = 't,'; // ث ARABIC LETTER THEH
var khahTo = 'k,'; // خ ARABIC LETTER KHAH
var thalTo = 'd,'; // ذ ARABIC LETTER THAL
var ghainTo = 'g,'; // غ ARABIC LETTER GHAIN
var maddaTo = ')o'; // آ ARABIC LETTER ALEF WITH MADDA ABOVE
var multiples = Object.freeze([thehTo, khahTo, thalTo, ghainTo, maddaTo]);
/**
 * @private
 * CAL destination writing
 * @const
 * @type { Writing }
 */
var calWriting = new Writing(
  Object.freeze(
    consonants.concat([
      thehTo, // '\u062B', ث ARABIC LETTER THEH
      khahTo, // '\u062E', خ ARABIC LETTER KHAH
      thalTo, // '\u0630', ذ ARABIC LETTER THAL
      'd', // '\u0636', ض ARABIC LETTER DAD - emphatic D missing in Aramaic
      'z', // '\u0638', ظ ARABIC LETTER ZAH - emphatic ðˤ, missing in aramaic
      ghainTo, // '\u063A', غ ARABIC LETTER GHAIN
      't', // '\u0629', ة ARABIC LETTER TEH MARBUTA
      ')', // '\u0621', ء Arabic Letter Hamza  - Garshuni hamzah
      maddaTo, // '\u0622', آ ARABIC LETTER ALEF WITH MADDA ABOVE
      ')', // '\u0623', أ ARABIC LETTER ALEF WITH HAMZA ABOVE
      'w', // '\u0624', ؤ ARABIC LETTER WAW WITH HAMZA ABOVE
      ')', // '\u0625', إ ARABIC LETTER ALEF WITH HAMZA BELOW
      'y' // '\u0626' ئ ARABIC LETTER YEH WITH HAMZA ABOVE
    ])
  ),
  Object.freeze([
    'a', //  َ Arabic fatha - Garshuni: a
    'o', //  ٰ Arabic letter superscript alef - Garshuni: long a
    'i', //  ِ Arabic kasra - Garshuni: i
    'u', //  ُ Arabic damma - Garshuni: u
    'a', //  ً Arabic fathatan - Garshuni: an
    'i', //  ٍ Arabic kasratan - Garshuni: in
    'u' //  ٌ Arabic dammatan - Garshuni: un
  ]),
  Object.freeze([
    '', //  ّ Arabic shadda - Garshuni
    '', //  ْ ARABIC SUKUN
    '', //  ٓ Arabic maddah above - Garshuni
    '', //  ٔ Arabic hamza above - Garshuni
    '' //  َ Arabic hamza below - Garshuni
  ]),
  Object.freeze([
    // using ; because , indicates spirantization
    ';', // ، Arabic COMMA - also used with Thaana and Syriac in modern text
    ';', // ؛ Arabic Semicolon - also used with Thaana and Syriac in modern text
    '?', // ؟ Arabic Question Mark - also used with Thaana and Syriac in modern text
    '!' // ! Exclamation Mark - regular ASCII exclamation mark
  ])
);

/**
 * @private
 * Maps input character to Arabic char
 * @param { string } c input character
 * @param { Object.<string, string> } fromTo mapping dictionary
 * @returns { string } Arabic mapped char
 */
var to = function (c, fromTo) { return fromTo[c] || (fromTo[c] === '' ? '' : c); };
var alef = '\u0627'; // ا ARABIC LETTER ALEF
var fatha = '\u064E'; //  َ Arabic fatha - Garshuni: a
var kasra = '\u0650'; //  ِ Arabic kasra - Garshuni: i
var damma = '\u064F'; //  ُ Arabic damma - Garshuni: u
var waw = '\u0648'; // و ARABIC LETTER WAW
var yeh = '\u064A'; // ي ARABIC LETTER YEH

/**
 * @private
 * Maps input character to CAL char
 * @param { string } c Arabic input character
 * @param { Object.<string, string> } fromTo mapping dictionary
 * @param { Object } wordProps optional word settings
 * @returns { string } CAL mapped char
 */
var mapCallback = function (word, i, fromTo, wordProps) {
  var c = word.charAt(i);
  if (!wordProps.isDotted) {
    return to(c, fromTo);
  }
  var m = '';
  var n = word.charAt(i + 1);
  switch (c) {
    case fatha:
      m =
        n === alef && word.length - 1 > i + 1 && word.charAt(i + 2) !== '-'
          ? '(o'
          : to(c, fromTo);
      break;
    case kasra:
      m =
        n === yeh
          ? 'yi' // Arabic stores as (iy)
          : to(c, fromTo);
      break;
    case damma:
      m =
        n === waw
          ? 'wu' // Arabic stores as (uw)
          : to(c, fromTo);
      break;
    default:
      m = to(c, fromTo);
      break;
  }
  return m;
};

/**
 * Arabic to CAL Mapper
 * @const
 * @type { Mapper }
 */
var mapper = new Mapper(arabicWriting, calWriting, mapCallback);
mapper.multiples = multiples;

/**
 * Convert from Arabic to CAL code
 * @static
 * @param {string} word input word in Arabic Unicode
 * @returns {string} the input word converted to CAL
 */
var toCal = function (word) {
  var wordProp = Object.freeze(
    Object.create(null, {
      isDotted: { value: isDotted(word), enumerable: true }
    })
  );
  return mapper.map(word, wordProp);
};

export { mapper, toCal };
